#ifndef LINT
static char *rcsid="$Id: ttyio.c,v 1.5 2000/04/01 21:26:40 crosser Exp $";
#endif

/*
	"ifmail": FidoNet(r) support package for UN*X platform

	Copyright (c) 1993-1999 Eugene G. Crosser

	You may distribute and/or use for any purpose modified or unmodified
	copies of this software if you preserve the copyright notice above.

	THIS SOFTWARE IS PROVIDED AS IS AND COME WITH NO WARRANTY OF ANY
	KIND, EITHER EXPRESSED OR IMPLIED.  IN NO EVENT WILL THE
	COPYRIGHT HOLDER BE LIABLE FOR ANY DAMAGES RESULTING FROM THE
	USE OF THIS SOFTWARE.

	URL: http://www.average.org/ifmail/
	E-Mail: <crosser@average.org>
*/

/*
	$Log: ttyio.c,v $
	Revision 1.5  2000/04/01 21:26:40  crosser
	experiments with Txy
	better treating of msgids/replyids
	
	Revision 1.4  1999/12/29 23:16:54  crosser
	make syslog and debug use %m for error message
	
	Revision 1.3  1999/12/28 16:01:54  crosser
	Move to syslog
	Portability for HP/UX
	
	Revision 1.2  1999/12/20 21:53:46  crosser
	Added Janus implementation by Kolya Nesterov, 2:463/567@fidonet
	(patch for 3.02 made by Vladimir Zarozhevsky)
	
	Revision 1.1.1.1  1999/08/29 22:53:46  crosser
	Ifmail
	
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/types.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <stdio.h>
#ifdef STDC_HEADERS
# include <stdlib.h>
# include <string.h>
#else
# ifdef HAVE_STRINGS_H
#  include <strings.h>
# endif
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#if defined(HAVE_TCP) || defined(HAVE_TERM)
#ifdef HAVE_SYS_BSDTYPES_H
#include <sys/bsdtypes.h>
#endif
#ifdef HAVE_NET_ERRNO_H
#include <net/errno.h>
#endif
#ifdef HAVE_NET_ERRNO_H
#include <net/errno.h>
#endif
#include <sys/socket.h>
#endif

#ifdef HAVE_SELECT
#ifdef HAVE_SYS_TIMES_H
#include <sys/times.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#else
#include <sys/time.h>
#endif
#else
#include <signal.h>
#endif

#ifdef HAVE_IOCTL_H
#include <sys/ioctl.h>
#endif

#ifdef NOISEDEBUG
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#endif

#include <termios.h>
#include "ttyio.h"
#include "lutil.h"

extern int hanged_up;

#ifndef BUFSIZ
#define BUFSIZ 1024
#endif

#ifndef NUMTIMERS
#define NUMTIMERS 3
#endif

int tty_status=0;
static char buffer[BUFSIZ];
static char *next;
static int left=0;

static time_t timer[NUMTIMERS];

	/* timer functions */

int tty_resettimern(tno)
int tno;
{
	if (tno >= NUMTIMERS)
	{
		errno=EINVAL;
		syslog(LOG_ERR,"invalid timer No for settimer()");
		return -1;
	}
	debug(18,"resetsettimer(%d)",tno);
	timer[tno]=(time_t)0;
	return 0;
}

int tty_running(tno)                                                           
int tno;                                                                       
{                                                                              
	/* check if timer is running */                                              
	if (timer[tno] == (time_t) 0)                                                
		return 0;                                                                  
	else                                                                         
		return 1;                                                                  
}                                                                              

void tty_resettimer(void)
{
	int i;

	debug(18,"resettimer");
	for (i=0;i<NUMTIMERS;i++) timer[i]=(time_t)0;
}

int tty_settimer(tno,interval)
int tno;
int interval;
{
	if (tno >= NUMTIMERS)
	{
		errno=EINVAL;
		syslog(LOG_ERR,"invalid timer No for settimer()");
		return -1;
	}
	debug(18,"settimer(%d,%d)",tno,interval);
	timer[tno]=time((time_t*)NULL)+interval;
	return 0;
}

int tty_expired(tno)
int tno;
{
	time_t now;

	(void)time(&now);
	if (tno >= NUMTIMERS)
	{
		errno=EINVAL;
		syslog(LOG_ERR,"invalid timer No for expired()");
		return -1;
	}
	debug(19,"expired(%d) now=%lu,timer=%lu,return %s",
		now,timer[tno],(now >= timer[tno])?"yes":"no");
	return (now >= timer[tno]);
}

	/* private r/w functions */

static int tty_read(buf,size,tot)
char *buf;
int size,tot;
{
	time_t timeout,now;
	int i,rc;

#ifdef HAVE_SELECT
	fd_set readfds,writefds,exceptfds;
	struct timeval seltimer;
#else
#error unimplemented non-select()-based i/o
#endif

	debug(18,"tty_read(%08lx,%d)",buf,size);
	if (size == 0) return 0;
	tty_status=0;

	(void)time(&now);
	timeout=(time_t)300; /* maximum of 5 minutes */
	for (i=0;i<NUMTIMERS;i++) if (timer[i])
	{
		if (now >= timer[i])
		{
			tty_status=STAT_TIMEOUT;
			debug(18,"timer %d already expired, return",i);
			return -tty_status;
		}
		else
		{
			if (timeout > (timer[i]-now))
				timeout=timer[i]-now;
		}
	}
	if ((tot != -1) && (timeout > tot))
		timeout=tot;
	debug(18,"tty_read timeout=%d",timeout);

#ifdef HAVE_SELECT
	FD_ZERO(&readfds);
	FD_ZERO(&writefds);
	FD_ZERO(&exceptfds);
	FD_SET(0,&readfds);
	FD_SET(0,&exceptfds);
	seltimer.tv_sec=timeout;
	seltimer.tv_usec=0;

	rc=select(1,&readfds,&writefds,&exceptfds,&seltimer);
	if (rc < 0)
	{
		if (hanged_up)
		{
			tty_status=STAT_HANGUP;
		}
		else
		{
			syslog(LOG_ERR,"select for read failed: %m");
			tty_status=STAT_ERROR;
		}
	}
	else if (rc == 0)
	{
		tty_status=STAT_TIMEOUT;
	}
	else /* rc > 0 */
	{
		if (FD_ISSET(0,&exceptfds)) tty_status=STAT_ERROR;
	}

	if (tty_status)
	{
		debug(18,"returning after select with status %d",tty_status);
		return -tty_status;
	}

	if (!FD_ISSET(0,&readfds))
	{
		syslog(LOG_ERR,"Cannot be: select returned but read fd not set");
		tty_status=STAT_ERROR;
		return -tty_status;
	}

	rc=read(0,buf,size);
	if (rc <= 0)
	{
		debug(18,"read() return %d",rc);
		if (hanged_up || (errno == EPIPE)
#if defined(HAVE_TCP) || defined(HAVE_TERM)
				|| (errno == ECONNRESET)
#endif
							)
			tty_status=STAT_HANGUP;
		else tty_status=STAT_ERROR;
		rc=-tty_status;
	}
	else debug(19,"read %d: \"%s\"",rc,printable(buf,rc));
	return rc;
#else
#endif
}

int tty_write(buf,size)
char *buf;
int size;
{
	int result;

	debug(18,"tty_write(%08lx,%d)",buf,size);
	tty_status=0;
	result=write(1,buf,size);
	if (result != size)
	{
		if (hanged_up || (errno == EPIPE)
#if defined(HAVE_TCP) || defined(HAVE_TERM)
				|| (errno == ECONNRESET)
#endif
							)
			tty_status=STAT_HANGUP;
		else tty_status=STAT_ERROR;
	}
	return -tty_status;
}

#define	WILL		251
#define	WONT		252
#define	DO		253
#define	DONT		254
#define	IAC		255

#define	TOPT_BIN	0
#define	TOPT_ECHO	1
#define	TOPT_SUPP	3

static void telnet_answer(tag,opt)
int tag,opt;
{
	char buf[3];

	buf[0]=IAC;
	buf[1]=tag;
	buf[2]=opt;
	tty_write(&buf,3);
}

void telnet_init(void)
{
	telnet_answer(DO,TOPT_SUPP);
	telnet_answer(WILL,TOPT_SUPP);
	telnet_answer(DO,TOPT_BIN);
	telnet_answer(WILL,TOPT_BIN);
	telnet_answer(DO,TOPT_ECHO);
	telnet_answer(WILL,TOPT_ECHO);
}

	/* public r/w functions */

int tty_check(void)
{
	if (left) return 1;
	left=tty_read(buffer,BUFSIZ,0);
	return (left > 0);
}

int tty_getc(tot)
int tot;
{
	if (!left)
	{
		left=tty_read(buffer,BUFSIZ,tot);
		next=buffer;
	}
	if (left <= 0)
	{
		left=0;
		return -tty_status;
	}
	else
	{
		left--;
		return (*next++)&0xff;
	}
}

int tty_get(buf,size,tot)
char *buf;
int size,tot;
{
	int result=0;

	if (left >= size)
	{
		memcpy(buf,next,size);
		next += size;
		left -= size;
		return 0;
	}

	if (left > 0)
	{
		memcpy(buf,next,left);
		buf += left;
		next += left;
		size -= left;
		left=0;
	}

	tty_resettimer();
	tty_settimer(0,tot);

	while ((result=tty_read(buf,size,-1)) > 0)
	{
		buf += result;
		size -= result;
	}

	tty_resettimer();

	return result;
}

int tty_putc(c)
int c;
{
	char buf=c;
	return tty_write(&buf,1);
}

int tty_put(buf,size)
char *buf;
int size;
{
	return tty_write(buf,size);
}

int tty_putget(obuf,osize,ibuf,isize)
char **obuf;
int *osize;
char **ibuf;
int *isize;
{
	time_t timeout,now;
	int i,rc;

#ifdef HAVE_SELECT
	fd_set readfds,writefds,exceptfds;
	struct timeval seltimer;
#else
#error unimplemented non-select()-based i/o
#endif

	tty_status=0;
	(void)time(&now);
	timeout=(time_t)300; /* maximum of 5 minutes */
	for (i=0;i<NUMTIMERS;i++) if (timer[i])
	{
		if (now >= timer[i])
		{
			tty_status=STAT_TIMEOUT;
			debug(18,"timer %d already expired, return",i);
			return -tty_status;
		}
		else
		{
			if (timeout > (timer[i]-now))
				timeout=timer[i]-now;
		}
	}
	debug(18,"tty_read timeout=%d",timeout);

#ifdef HAVE_SELECT
	FD_ZERO(&readfds);
	FD_ZERO(&writefds);
	FD_ZERO(&exceptfds);
	FD_SET(0,&readfds);
	FD_SET(1,&writefds);
	FD_SET(0,&exceptfds);
	FD_SET(1,&exceptfds);
	seltimer.tv_sec=timeout;
	seltimer.tv_usec=0;

	rc=select(2,&readfds,&writefds,&exceptfds,&seltimer);
	if (rc < 0)
	{
		if (hanged_up)
		{
			tty_status=STAT_HANGUP;
		}
		else
		{
			syslog(LOG_ERR,"select for putget failed: %m");
			tty_status=STAT_ERROR;
		}
	}
	else if (rc == 0)
	{
		tty_status=STAT_TIMEOUT;
	}
	else /* rc > 0 */
	{
		if ((FD_ISSET(0,&exceptfds)) || (FD_ISSET(1,&exceptfds)))
			tty_status=STAT_ERROR;
	}

	if (tty_status)
	{
		debug(18,"returning after select with status %d",tty_status);
		return -tty_status;
	}

	if (FD_ISSET(0,&readfds) && *isize)
	{
		rc=read(0,*ibuf,*isize);
		if (rc < 0)
		{
			syslog(LOG_ERR,"read failed: %m");
			tty_status=STAT_ERROR;
		}
		else
		{
			(*ibuf)+=rc;
			(*isize)-=rc;
		}
	}
	if (FD_ISSET(1,&writefds) && *osize)
	{
		rc=write(1,*obuf,*osize);
		if (rc < 0)
		{
			syslog(LOG_ERR,"write failed: %m");
			tty_status=STAT_ERROR;
		}
		else
		{
			(*obuf)+=rc;
			(*osize)-=rc;
		}
	}
#else
#endif

	if (tty_status) return -tty_status;
	else return ((*isize == 0) | ((*osize == 0) << 1));
}

void tty_flushin(void)
{
	tcflush(0, TCIFLUSH);
}

void tty_flushout(void)
{
	tcflush(1, TCOFLUSH);
}
