#ifndef LINT
static char *rcsid="$Id: areas.c,v 1.3 1999/12/29 23:16:54 crosser Exp $";
#endif

/*
	"ifmail": FidoNet(r) support package for UN*X platform

	Copyright (c) 1993-1999 Eugene G. Crosser

	You may distribute and/or use for any purpose modified or unmodified
	copies of this software if you preserve the copyright notice above.

	THIS SOFTWARE IS PROVIDED AS IS AND COME WITH NO WARRANTY OF ANY
	KIND, EITHER EXPRESSED OR IMPLIED.  IN NO EVENT WILL THE
	COPYRIGHT HOLDER BE LIABLE FOR ANY DAMAGES RESULTING FROM THE
	USE OF THIS SOFTWARE.

	URL: http://www.average.org/ifmail/
	E-Mail: <crosser@average.org>
*/

/*
	$Log: areas.c,v $
	Revision 1.3  1999/12/29 23:16:54  crosser
	make syslog and debug use %m for error message
	
	Revision 1.2  1999/12/28 16:01:54  crosser
	Move to syslog
	Portability for HP/UX
	
	Revision 1.1.1.1  1999/08/29 22:53:45  crosser
	Ifmail
	
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/types.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <stdio.h>
#ifdef STDC_HEADERS
# include <stdlib.h>
# include <string.h>
#else
# ifdef HAVE_STRINGS_H
#  include <strings.h>
# endif
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "lutil.h"
#include "xutil.h"
#include "areas.h"
#include "rdconfig.h"
#include "lhash.h"

typedef struct _areang {
	char *area;
	char *ng;
	char *dist;
} areang;

extern area_list *badgroups;
static LHASH *nglist_byarea=(LHASH*)0;
static LHASH *nglist_byng=(LHASH*)0;
char *defgroup=NULL,*defdist=NULL,*defarea=NULL;
int group_count;

static char *expand(p,q)
char *p,*q;
{
	char *r,*tmp,*ret;

	debug(9,"expanding \"%s\" + \"%s\"",S(p),S(q));

	tmp=xstrcpy(p);
	if ((r=strchr(tmp,'*'))) *r++='\0';
	ret=xstrcpy(tmp);
	if (r)
	{
		ret=xstrcat(ret,q);
		ret=xstrcat(ret,r);
	}
	free(tmp);
	debug(9,"expand result \"%s\"",S(ret));
	return ret;
}

static unsigned long hash_byarea(hash_entry)
areang *hash_entry;
{
	return lh_strhash(hash_entry->area);
}

static unsigned long hash_byng(hash_entry)
areang *hash_entry;
{
	return lh_strhash(hash_entry->ng);
}

static int cmp_byarea(e1,e2)
areang *e1,*e2;
{
	return strcmp(e1->area,e2->area);
}

static int cmp_byng(e1,e2)
areang *e1,*e2;
{
	return strcmp(e1->ng,e2->ng);
}

void tidy_arealist(al)
area_list *al;
{
	area_list *tmp;

	for (;al;al=tmp)
	{
		tmp=al->next;
		if (al->name) free(al->name);
		free(al);
	}
}

area_list *areas(ngroups)
char *ngroups;
{
	char *ng,*p,*q,*r=NULL;
	area_list *start=NULL,**cur,*bg;
	areang *ngl;
	areang dummy_entry;
	int forbid=0;

	if (ngroups == NULL) return NULL;
	ng=xstrcpy(ngroups);
	cur=&start;

	group_count=0;
	for (p=strtok(ng," ,\t\n");p;p=strtok(NULL," ,\t\n"))
	{
		group_count++;
		q=NULL;
		for (bg=badgroups;bg;bg=bg->next)
		if (!strncmp(bg->name,p,strlen(bg->name)))
		{
			syslog(LOG_INFO,"Message not gated due to forbidden group %s",
				S(p));
			forbid=1;
			break;
		}
		dummy_entry.ng=p;
		if ((ngl=(areang*)lh_retrieve(nglist_byng,(char*)&dummy_entry)))
		{
			q=ngl->area;
		}
		else q=NULL;
		if (q || defarea)
		{
			if (q == NULL)
			{
				for (r=p;*r;r++) *r=toupper(*r);
				q=expand(defarea,p);
			}
			debug(9,"newsgroup \"%s\" -> area \"%s\"",S(p),S(q));
			(*cur)=(area_list *)xmalloc(sizeof(area_list));
			(*cur)->next=NULL;
			(*cur)->name=xstrcpy(q);
			cur=&((*cur)->next);
		}
	}

	free(ng);
	if (forbid)
	{
		tidy_arealist(start);
		return NULL;
	}
	else return start;
}

void ngdist(area,newsgroup,distribution)
char *area,**newsgroup,**distribution;
{
	areang *ang;
	areang dummy_entry;
	char *p,*q;

	*newsgroup=NULL;
	*distribution=NULL;
	p=xstrcpy(area);
	if ((*(q=p+strlen(p)-1) == '\r') || (*q == '\n')) *q='\0';

	dummy_entry.area=p;
	if ((ang=(areang*)lh_retrieve(nglist_byarea,(char*)&dummy_entry)))
	{
		*newsgroup=ang->ng;
		*distribution=ang->dist;
		debug(9,"area \"%s\" -> newsgroup \"%s\", dist \"%s\"",
			S(p),S(*newsgroup),S(*distribution));
	}
	else
	{
		*newsgroup = NULL;
		*distribution = NULL;
	}

	if ((*newsgroup == NULL) && defgroup)
	{
		for (q=p;*q;q++) *q=tolower(*q);
		*newsgroup=expand(defgroup,p);
		*distribution=defdist;
		debug(9,"area \"%s\" -> default newsgroup \"%s\", dist \"%s\"",
			S(p),S(*newsgroup),S(*distribution));
	}

	if (*newsgroup == NULL)
	{
		syslog(LOG_INFO,"No newsgroup for area tag %s",S(p));
	}

	free(p);
}

void readareas(fn)
char *fn;
{
	FILE *fp;
	char buf[128],*p,*q,*r;
	areang *cur;

	if (defgroup) free(defgroup);
	defgroup=NULL;
	if (defdist) free(defdist);
	defdist=NULL;
	if (defarea) free(defarea);
	defarea=NULL;

	if (!nglist_byarea) nglist_byarea=lh_new(hash_byarea,cmp_byarea);
	if (!nglist_byng) nglist_byng=lh_new(hash_byng,cmp_byng);

	if ((fp=fopen(fn,"r")) == NULL)
	{
		syslog(LOG_ERR,"unable to open area file %s: %m",S(fn));
		return;
	}
	while (fgets(buf,sizeof(buf)-1,fp))
	{
		if (isspace(buf[0])) continue;
		p=strtok(buf," \t\n");
		q=strtok(NULL," \t\n");
		r=strtok(NULL," \t\n");
		if (p && q)
		{
			if (strcmp(q,"*") == 0)
			{
				debug(9,"adding default areatag \"%s\"",S(p));
				defarea=xstrcpy(p);
			}
			if (strcmp(p,"*") == 0)
			{
				debug(9,"adding default group \"%s\" (dist %s)",
					S(q),S(r));
				defgroup=xstrcpy(q);
				defdist=xstrcpy(r);
			}
			if ((strcmp(p,"*") != 0) && (strcmp(q,"*") != 0))
			{
				debug(9,"adding area \"%s\" for ng \"%s\" (dist %s)",
					S(p),S(q),S(r));
				cur=(areang*)malloc(sizeof(areang));
				cur->area=xstrcpy(p);
				cur->ng=xstrcpy(q);
				if (r) cur->dist=xstrcpy(r);
				else cur->dist=NULL;
				lh_insert(nglist_byarea,(char*)cur);
				lh_insert(nglist_byng,(char*)cur);
			}
		}
	}
	fclose(fp);
}
