#ifndef LINT
static char *rcsid="$Id: lutil.c,v 1.7 2000/04/15 09:46:46 crosser Exp $";
#endif

/*
	"ifmail": FidoNet(r) support package for UN*X platform

	Copyright (c) 1993-1999 Eugene G. Crosser

	You may distribute and/or use for any purpose modified or unmodified
	copies of this software if you preserve the copyright notice above.

	THIS SOFTWARE IS PROVIDED AS IS AND COME WITH NO WARRANTY OF ANY
	KIND, EITHER EXPRESSED OR IMPLIED.  IN NO EVENT WILL THE
	COPYRIGHT HOLDER BE LIABLE FOR ANY DAMAGES RESULTING FROM THE
	USE OF THIS SOFTWARE.

	URL: http://www.average.org/ifmail/
	E-Mail: <crosser@average.org>
*/

/*
	$Log: lutil.c,v $
	Revision 1.7  2000/04/15 09:46:46  crosser
	fix arguments in replacement syslog
	
	Revision 1.6  2000/04/10 20:44:37  crosser
	make it compile without syslog
	
	Revision 1.5  1999/12/29 23:16:54  crosser
	make syslog and debug use %m for error message
	
	Revision 1.4  1999/12/28 23:34:27  crosser
	cleaning up
	
	Revision 1.3  1999/12/28 16:01:55  crosser
	Move to syslog
	Portability for HP/UX
	
	Revision 1.2  1999/08/30 16:43:16  crosser
	Make the thing at least compile.
	
	Revision 1.1.1.1  1999/08/29 22:53:45  crosser
	Ifmail
	
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/types.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <stdio.h>
#ifdef STDC_HEADERS
# include <stdlib.h>
# include <string.h>
#else
# ifdef HAVE_STRINGS_H
#  include <strings.h>
# endif
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <sys/stat.h>
#include <varargs.h>
#include <errno.h>
#ifdef HAVE_SYSLOG
#include <syslog.h>
#endif
#include "xutil.h"
#include "lutil.h"

#ifdef HAVE_SYSLOG
static int syslog_opened=0;
#else
FILE *logfile=NULL;
FILE *errfile=NULL;
#endif
int logfacility=LOG_USER;
char *myname="unknown";
char *log_name="/tmp/tmplog";
char *errname="/tmp/tmplog";
unsigned long verbose=0L;

void setmyname(arg)
char *arg;
{
	if ((myname=strrchr(arg,'/'))) myname++;
	else myname=arg;
	myname=xstrcpy(myname);
#ifdef HAVE_SYSLOG
	if (syslog_opened) closelog();
	openlog(myname,LOG_PID|LOG_CONS,logfacility);
	syslog_opened=1;
#endif
}

void setnamesfx(arg)
char *arg;
{
	char *p;

	if (arg == NULL) return;

	if ((p=strchr(myname,'.'))) *p='\0';
	myname=xstrcat(myname,".");
	myname=xstrcat(myname,arg);
#ifdef HAVE_SYSLOG
	if (syslog_opened) closelog();
	openlog(myname,LOG_PID|LOG_CONS,logfacility);
	syslog_opened=1;
#endif
}

#ifndef HAVE_STRERROR
char *strerror(errno)
int errno;
{
#ifndef HAVE_SYS_ERRLIST
	extern char *sys_errlist[];
#endif
	extern int sys_nerr;
	if ((errno >= 0) && (errno <= sys_nerr))
		return(sys_errlist[errno]);
	else
		return("unknown error");
}
#endif

static char *mon[] = {
"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"
};

char *date(t)
time_t t;
{
	struct tm ptm;
	time_t now;
	static char buf[20];

	if (t) now=t; 
	else time(&now);
	ptm=*localtime(&now);
	sprintf(buf,"%s %02d %02d:%02d:%02d",
		mon[ptm.tm_mon],ptm.tm_mday,
		ptm.tm_hour,ptm.tm_min,ptm.tm_sec);
	return(buf);
}

static char *pbuff = NULL;

char *printable(s,l)
char *s;
int l;
{
	int len;
	char *p;

	if (pbuff) free(pbuff);
	pbuff=NULL;

	if (s == NULL) return "(null)";

	if (l > 0) len=l;
	else if (l == 0) len=strlen(s);
	else {
		len=strlen(s);
		if (len > -l) len=-l;
	}
	pbuff=(char*)malloc(len*4+1);
	p=pbuff;
	while (len--) {
		if (*(unsigned char*)s >= ' ') *p++=*s;
		else switch (*s) {
		case '\\': *p++='\\'; *p++='\\'; break;
		case '\r': *p++='\\'; *p++='r'; break;
		case '\n': *p++='\\'; *p++='n'; break;
		case '\t': *p++='\\'; *p++='t'; break;
		case '\b': *p++='\\'; *p++='b'; break;
		default:   sprintf(p,"\\%03o",*s); p+=4; break;
		}
		s++;
	}
	*p='\0';
	return pbuff;
}

char *printablec(c)
char c;
{
	return printable(&c,1);
}

		/* log and debug outputs */

#ifndef HAVE_SYSLOG
void openlog(nm,opt,fac)
char *nm;
int opt;
int fac;
{
	int oldmask;

	if (!logfile) {
		oldmask=umask(066);
		logfile=fopen(nm,"a");
		umask(oldmask);
		if (logfile == NULL) {
			perror("Cannot open log file");
			logfile=fdopen(2,"a");
		}
	}
}

void closelog(void)
{
	if (logfile) fclose(logfile);
	logfile=NULL;
}

void syslog(va_alist)
va_dcl
{
	va_list	args;
	char	*fmt;
	char	fmt_cpy[1024];
	int	saved_errno=errno;
	int	flag; /* unused */

	va_start(args);
	flag=va_arg(args,int);
	fmt=va_arg(args,char*);
	if (!logfile) openlog(log_name,0,LOG_USER);

	/* substitute error message for %m */
	{
		register char ch, *t1, *t2;

		for (t1 = fmt_cpy;
		     (ch = *fmt) != '\0' && t1<fmt_cpy+sizeof(fmt_cpy);
		     ++fmt)
			if (ch == '%' && fmt[1] == 'm') {
				++fmt;
				for (t2 = strerror(saved_errno);
				     (*t1 = *t2++); ++t1);
			} else *t1++ = ch;
		*t1 = '\0';
	}

	fprintf(logfile,"%s local %s[%d]: ",date(0L),myname,getpid());
	vfprintf(logfile,fmt_cpy,args);
	fprintf(logfile,"\n");
	fflush(logfile);

	va_end(args);
	return;
}
#endif

void debug(va_alist)
va_dcl
{
	va_list	args;
	unsigned long	level;
	char	*fmt;
	char	fmt_cpy[1024];
	int	saved_errno=errno;

	va_start(args);
	level=va_arg(args, unsigned long);
	fmt=va_arg(args, char*);

	/* substitute error message for %m */
	{
		register char ch, *t1, *t2;

		for (t1 = fmt_cpy;
		     (ch = *fmt) != '\0' && t1<fmt_cpy+sizeof(fmt_cpy);
		     ++fmt)
			if (ch == '%' && fmt[1] == 'm') {
				++fmt;
				for (t2 = strerror(saved_errno);
				     (*t1 = *t2++); ++t1);
			} else *t1++ = ch;
		*t1 = '\0';
	}

	if ((verbose && (level == 0)) || (verbose & (1 << (level-1)))) {
		vfprintf(stderr,fmt_cpy,args);
		fprintf(stderr,"\n");
		fflush(stderr);
	}
	va_end(args);
	return;
}

unsigned long setverbose(s)
char *s;
{
	int shift;
	unsigned long tv=0L;
	int bad=0;

	if (strlen(s) == strspn(s,"0123456789"))
	{
		shift=atoi(s);
		if ((shift > 0) && (shift <= 32))
			while (--shift >= 0)
				tv |= 1<<shift;
		else if (shift != 0) bad=1;
	}
	else while ((shift=*s++))
	{
		if ((shift >= 'A') && (shift <= 'Z'))
			shift += ('a'-'A');
		if ((shift >= 'a') && (shift <= 'z'))
			tv |= 1<<(shift -'a');
		else bad=1;
	}

	if (tv && ((shift=getuid()) != 0) && (shift != geteuid()))
	{
		fprintf(stderr,"Debug request denied, run as user id %ld\n",
			(long)geteuid());
		return 0L;
	}

	if (bad) fprintf(stderr,"invalid debug verbosity code\n");

	return tv;
}
